## Spectral Regression Discriminant Analysis.

## This is an implementation of Spectral Regression Discriminant Analysis described in:
## 'SRDA: An Efficient Algorithm for Large ScaleDiscriminant Analysis' Deng Cai,
## Xiaofei He, Jiawei Han. 2008.

## This code is written by Roberto Visintainer, <visintainer@fbk.eu> and Davide Albanese, <albanese@fbk.eu>.
## (C) 2008 Fondazione Bruno Kessler - Via Santa Croce 77, 38100 Trento, ITALY.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


__all__ = ['Srda']

from numpy import *
from numpy.linalg import inv


class Srda:
    """Spectral Regression Discriminant Analysis (SRDA).

    Example:
    
    >>> from numpy import *
    >>> from mlpy import *
    >>> xtr = array([[1.0, 2.0, 3.1, 1.0],  # first sample
    ...              [1.0, 2.0, 3.0, 2.0],  # second sample
    ...              [1.0, 2.0, 3.2, 1.0]]) # third sample
    >>> ytr = array([1, -1, 1])             # classes
    >>> mysrda = Srda()                     # initialize srda class
    >>> mysrda.compute(xtr, ytr)            # compute srda
    1                                      
    >>> mysrda.predict(xtr)                 # predict srda model on training data
    array([ 1, -1,  1])
    >>> xts = array([4.0, 5.0, 6.0, 7.0])   # test point
    >>> mysrda.predict(xts)                 # predict srda model on test point
    -1
    >>> mysrda.realpred                     # real-valued prediction
    -16.500000000001439
    >>> mysrda.weights(xtr, ytr)            # compute weights on training data
    array([  1.00000000e+00,   2.00000000e+00,   5.40012479e-13,
             4.50000000e+00])
    """

    def __init__ (self, alpha = 1.0):
        """Initialize the Srda class.

        Input
        
          * *alpha* - [float] (>=0.0) regularization parameter
        """

        if alpha < 0.0:
            raise ValueError("alpha (regularization parameter) must be >= 0.0")
        
        self.__alpha = alpha
        
        self.__classes = None
        self.__a = None
        self.__th = 0.0
        self.__computed = False
        
        self.realpred = None
        
    def compute (self, x, y):
        """Compute Srda model.

        Initialize array of alphas a.
       
        Input
        
          * *x* - [2D numpy array float] (sample x feature) training data
          * *y* - [1D numpy array integer] (two classes) classes

        Output
        
          * 1
        """

        # See eq 19 and 24
        
        self.__classes = unique(y)
        if self.__classes.shape[0] != 2:
            raise ValueError("SRDA works only for two-classes problems")

        cl0 = where(y == self.__classes[0])[0]
        cl1 = where(y == self.__classes[1])[0]             
        ncl0 = cl0.shape[0]
        ncl1 = cl1.shape[0]
        
        y0 = x.shape[0] / float(ncl0)
        y1 = -x.shape[0] / float(ncl1)              

        ym = append(ones(ncl0) * y0, ones(ncl1) * y1, axis = 1)

        newpos = r_[cl0, cl1]
        
        xi = x[newpos].copy()
        xi = append(xi, ones(x.shape[0]).reshape(x.shape[0], -1), axis = 1)
        xiT = xi.transpose()

        xXI = inv(dot(xi, xiT) + (self.__alpha * identity(x.shape[0])))
                
        c = dot(xXI, ym)
        self.__a = dot(xiT, c)

        ##### Threshold tuning ######
        ncomptrue = empty(x.shape[0], dtype = int)
        ths       = empty(x.shape[0])
        ytmp      = empty_like(y)

        self.__computed = True

        self.predict(x)
        rpsorted = sort(self.realpred)
        
        for t in range(ths.shape[0] - 1):
            ths[t] = (rpsorted[t] + rpsorted[t + 1]) * 0.5         
            ytmp[self.realpred <= ths[t]] = self.__classes[0]
            ytmp[self.realpred >  ths[t]] = self.__classes[1]
            comp = (y == ytmp)
            ncomptrue[t] = sum(comp)

        # Try th = 0.0
        ths[-1] = 0.0
        ytmp[self.realpred <=  ths[-1]] = self.__classes[0]
        ytmp[self.realpred > ths[-1]] = self.__classes[1]
        comp = (y == ytmp)
        ncomptrue[-1] = sum(comp)

        self.__th = ths[argmax(ncomptrue)]          
        #############################
                
        return 1
    
    def weights (self, x, y):
        """Return feature weights.
        
        Input

          * *x* - [2D numpy array float] (sample x feature) training data
          * *y* - [1D numpy array integer] (two classes) classes

        Output
        
          * *fw* - [1D numpy array float] feature weights
        """

        self.compute(x, y)
        return abs(self.__a[0:x.shape[1]])

    def predict (self, p):
        """Predict Srda model on test point(s).

        Input
        
          * *p* - test point(s) [1D or 2D numpy array float]
        
        Output
        
          * *cl* - [integer or 1D numpy array integer] class(es) predicted
          * *self.realpred* - [float or 1D numpy array float] real valued prediction
        """
        
        if self.__computed == False:
            raise StandardError("No SRDA model computed")
        
        if p.ndim == 2:
            p = append(p, ones(p.shape[0]).reshape(p.shape[0], -1), axis = 1)
            yresult = empty((p.shape[0]), int)
            cl = -dot(self.__a, p.transpose())
            self.realpred = cl
            
            yresult[cl <= self.__th] = self.__classes[0]
            yresult[cl >  self.__th] = self.__classes[1]

            return yresult

        elif p.ndim == 1:
            p = r_[p, 1]

            cl = -dot(p, self.__a)
            self.realpred = cl
                                
            if  cl <= self.__th:
                yonecamp = self.__classes[0]
            elif cl > self.__th:
                yonecamp = self.__classes[1]

            return yonecamp
        
