## This file is part of mlpy.
## Support Vector Machines (SVM) based on SVM
## C-libraries developed by Stefano Merler.

## For feature weights see:
## C. Furlanello, M. Serafini, S. Merler, and G. Jurman.
## Advances in Neural Network Research: IJCNN 2003.
## An accelerated procedure for recursive feature ranking
## on microarray data.
## Elsevier, 2003.
    
## This code is written by Davide Albanese, <albanese@fbk.eu>.
## (C) 2007 Fondazione Bruno Kessler - Via Santa Croce 77, 38100 Trento, ITALY.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['Svm']

from numpy import *
import svmcore


def KernelGaussian (x1, x2, kp):
    """
    Gaussian kernel
    K(x1,x2,kp) = exp^(-(||x1-x2||)^2 / kp)
    """
    
    sub = x1 - x2
    norm = (sum(abs(sub)**2))**(0.5)
    return exp(-norm**2 / kp)

def MatrixKernelGaussian(x, kp):
    """
    Create the matrix K
    K[i, j] = KernelGaussian(x[i], x[j], kp)
    """
    
    K = empty((x.shape[0], x.shape[0]))
    for i in xrange(x.shape[0]):
        for j in xrange(i, x.shape[0]):
            K[i,j] = KernelGaussian(x[i], x[j], kp)
            K[j,i] = K[i,j]          
    return K

def computeZ(K, y):
  """
  Compute the matrix Z[i,j] = y[i]*y[j]*K[i,j]

  See Maria Serafini Thesis, p. 29.
  """
  
  Z = empty((y.shape[0], y.shape[0]))     
  for i in xrange(y.shape[0]):
      for j in xrange(y.shape[0]):
          Z[i, j] = y[i] * y[j] * K[i, j]
          return Z

def computeZ_k(Z, k, x, kp):
  """
  Compute Z_k[i, j] = Z[i, j] * exp((x[i, k]-x[j, k])**2 / kp)

  See Maria Serafini Thesis, p. 30.
  """
  
  Z_k = empty_like(Z)  
  for i in xrange(Z.shape[0]):
      for j in xrange(Z.shape[0]):
          e = exp((x[i, k]-x[j, k])**2 / kp)
          if abs(e) == inf:
              raise StandardError("kp is too small or the data is not standardized")
          Z_k[i, j] = Z[i, j] * e
  return Z_k


class Svm:
    """
    Support Vector Machines (SVM).

    Example:
    
    >>> from numpy import *
    >>> from mlpy import *
    >>> xtr = array([[1.0, 2.0, 3.0, 1.0],  # first sample
    ...              [1.0, 2.0, 3.0, 2.0],  # second sample
    ...              [1.0, 2.0, 3.0, 1.0]]) # third sample
    >>> ytr = array([1, -1, 1])             # classes
    >>> mysvm = Svm()                       # initialize Svm class
    >>> mysvm.compute(xtr, ytr)             # compute SVM
    1
    >>> mysvm.predict(xtr)                  # predict SVM model on training data
    array([ 1, -1,  1])
    >>> xts = array([4.0, 5.0, 6.0, 7.0])   # test point
    >>> mysvm.predict(xts)                  # predict SVM model on test point
    -1
    >>> mysvm.realpred                      # real-valued prediction
    -5.5
    >>> mysvm.weights(xtr, ytr)             # compute weights on training data
    array([ 0.,  0.,  0.,  1.])
    """
    
    def __init__(self, kernel = 'linear', kp = 0.1, C = 1.0, tol = 0.001,
                 eps = 0.001, maxloops = 1000, cost = 0.0):
        """Initialize the Svm class.
        
        Input
        
          * *kernel*    - [string] kernel ('linear', 'gaussian', 'polynomial', 'tr')
          * *kp*        - [float] kernel parameter (two sigma squared) for gaussian and polynomial kernel
          * *C*         - [float] regularization parameter
          * *tol*       - [float] tolerance for testing KKT conditions
          * *eps*       - [float] convergence parameter
          * *maxloops*  - [integer] maximum number of optimization loops
          * *cost*      - [float] for cost-sensitive classification (-1.0, ..., 1.0)
        """

        SVM_KERNELS = {'linear': 1,
                       'gaussian': 2,
                       'polynomial': 3,
                       'tr': 4}
        
        self.__kernel   = SVM_KERNELS[kernel]
        self.__kp       = kp
        self.__C        = C
        self.__tol      = tol
        self.__eps      = eps
        self.__maxloops = maxloops
        self.__cost     = cost

        self.__x    = None
        self.__y    = None
        self.__w    = None
        self.__a    = None
        self.__b    = None
        self.__conv = None # svm convergence
        self.realpred = None

        self.__computed = False

        # For 'terminated ramps' (tr) only
        self.__sf_w  = None
        self.__sf_b  = None
        self.__sf_i  = None
        self.__sf_j  = None
        self.__nsf   = None
        self.__svm_x = None
        ################################     

    def compute(self, x, y):
        """Compute Svm model.

        Input
        
          * *x* - [2D numpy array float] (sample x feature) training data
          * *y* - [1D numpy array integer] (-1 or 1) classes

        Output
        
          * *conv* - [integer] svm convergence (0: false, 1: true)
        """

        classes = unique(y)
        if classes.shape[0] != 2:
            raise ValueError("Svm works only for two-classes problems")
     
        # Store x and y
        self.__x = x.copy()
        self.__y = y.copy()
       
        # Kernel 'tr'
        if self.__kernel == 4: 
            res = svmcore.computesvmtr(self.__x, self.__y, self.__C, self.__tol,
                                       self.__eps, self.__maxloops, self.__cost)
            self.__w        = res[0]
            self.__a        = res[1]
            self.__b        = res[2]
            self.__conv     = res[3]
            self.__sf_w     = res[4]
            self.__sf_b     = res[5]
            self.__sf_i     = res[6]
            self.__sf_j     = res[7]
            self.__svm_x    = res[8]
            self.__nsf      = self.__sf_w.shape[0]
            self.__computed = True
            
        # Kernel 'linear', 'gaussian', 'polynomial'
        else:
            res = svmcore.computesvm(self.__x, self.__y, self.__kernel, self.__kp, self.__C,
                                     self.__tol, self.__eps, self.__maxloops, self.__cost)
            
            self.__w = res[0]
            self.__a = res[1]
            self.__b = res[2]
            self.__conv = res[3]
            self.__computed = True

        # Return convergence
        return self.__conv
    
    def predict(self, p):
        """Predict svm model on a test point(s).

        Input
        
          * *p* - [1D or 2D numpy array float] test point(s)

        Output
        
          * *cl* - [integer or 1D numpy array integer] class(es) predicted
          * *self.realpred* - [1D numpy array float] real valued prediction
        """
        
        if self.__computed == False:
            raise StandardError("No SVM model computed")

        # Kernel 'tr'
        if self.__kernel == 4:
            if p.ndim == 1:
                self.realpred = svmcore.predictsvmtr(self.__x, self.__y, p, self.__w, self.__b,
                                                     self.__sf_w, self.__sf_b, self.__sf_i, self.__sf_j)
            elif p.ndim == 2:
                self.realpred = empty(p.shape[0], dtype = float)
                for i in range(p.shape[0]):
                    self.realpred[i] = svmcore.predictsvmtr(self.__x, self.__y, p[i], self.__w, self.__b,
                                                            self.__sf_w, self.__sf_b, self.__sf_i, self.__sf_j)
        # Kernel 'linear', 'gaussian', 'polynomial'
        else:
            if p.ndim == 1:
                self.realpred = svmcore.predictsvm(self.__x, self.__y, p, self.__w, self.__a,
                                                   self.__b, self.__kp, self.__kernel)
            elif p.ndim == 2:
                self.realpred = empty(p.shape[0], dtype = float)
                for i in range(p.shape[0]):
                    self.realpred[i] = svmcore.predictsvm(self.__x, self.__y, p[i], self.__w, self.__a,
                                                          self.__b, self.__kp, self.__kernel)            
        # Return prediction
        if p.ndim == 1:
            pred = 0
            if self.realpred > 0.0:
                pred = 1
            elif self.realpred < 0.0:
                pred = -1

        if p.ndim == 2:
            pred = zeros(p.shape[0], dtype = int)
            pred[where(self.realpred > 0.0)[0]] = 1
            pred[where(self.realpred < 0.0)[0]] = -1
                   
        return pred

    def weights(self, x, y):
        """Return feature weights.

        Input
        
          * *x* - [2D numpy array float] (sample x feature) training data
          * *y* - [1D numpy array integer] (-1 or 1) classes

        Output
        
          * *fw* - [1D numpy array float] feature weights
        """

        self.compute(x, y)

        # Linear case
        if self.__kernel == 1:
            return self.__w**2

        # Gaussian and polynomial case
        elif self.__kernel in [2,3]:
            K = MatrixKernelGaussian(self.__x, self.__kp)
            Z = computeZ(K, self.__y)          
            # Compute dJ[i] = 0.5*a*Z*aT - 0.5*a*Z*(-i)*aT
            a = self.__a
            aT = self.__a.reshape(-1,1)
            dJ1 = 0.5 * dot(dot(a, Z), aT)
            dJ2 = empty(self.__x.shape[1])
            for i in range(self.__x.shape[1]):
                Z_i = computeZ_k(Z, i, self.__x, self.__kp) # Compute Z_i                
                dJ2[i] = 0.5 * dot(dot(a, Z_i), aT)
            return dJ1 - dJ2

        # Tr case
        elif self.__kernel == 4:

            w      = empty((self.__nsf, x.shape[1]))
            norm_w = zeros((self.__nsf,))
            a      = zeros((self.__nsf,))
            h      = zeros((x.shape[1],))
                       
            for t in range(self.__nsf):
                it = self.__sf_i[t]
                jt = self.__sf_j[t]

                w[t] = abs(self.__sf_w[t] * (self.__y[it] * self.__x[it] + self.__y[jt] * self.__x[jt]))
                norm_w[t] = abs(w[t]).sum()

                for i in range(x.shape[0]):
                    a[t] += self.__y[i] * self.__a[i] * self.__svm_x[i][t]

            for j in range(x.shape[1]):
                for t in range(self.__nsf):
                    h[j] += abs(a[t]) * w[t][j] / norm_w[t]

            return h
        
